//======================================================================
//
// K - Beispieltreiber (layered driver)
//
// legt sich ueber den Standard-Tastaturtreiber und simuliert die
// Tastenkombination Ctrl+Alt+Entf bei Druck auf F12
//
//======================================================================

#include "ntddk.h"
#include "stdarg.h"
#include "stdio.h"
#include "k.h"
#include "ntddkbd.h"
#include "k_msg.h"

//
// Die Uebersetzung der Debugmeldungen (DbgPrint) soll ebenso wie
// feste Breakpoints (DbgBreak) nur im "Checked-Build" erfolgen.
//
#if DBG
 #define DbgPrint(arg) DbgPrint arg
 #define DbgBreak() DbgBreakPoint()
#else
 #define DbgPrint(arg)
 #define DbgBreak()
#endif


//----------------------------------------------------------------------
//                         Globale Vereinbarungen
//---------------------------------------------------------------------- 
PDEVICE_OBJECT       kbdDevice;
PDEVICE_OBJECT       HookDeviceObject;

//----------------------------------------------------------------------
//                         Forward Definitionen
//---------------------------------------------------------------------- 
NTSTATUS K_DispatchRead( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );
NTSTATUS K_DispatchGeneral(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp );
NTSTATUS K_Init( IN PDRIVER_OBJECT DriverObject );
BOOLEAN K_ReportEvent(IN NTSTATUS ErrorCode,IN ULONG UniqueErrorValue,
                      IN PVOID IoObject,IN PIRP Irp,PUCHAR InsertString,
                      IN ULONG StringSize);

//----------------------------------------------------------------------
//
// DriverEntry - Standardfunktion fuer den Treibereintrittspunkt
//
// Aufgaben:
//            - Initialisierung der Datenstrukturen
//            - [...]
//
//----------------------------------------------------------------------
NTSTATUS DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath )
{
    PDEVICE_OBJECT         DeviceObject        = NULL;
    NTSTATUS               mStatus;
    WCHAR                  Name[]  = L"\\Device\\K";
    UNICODE_STRING         uniName;
    WCHAR                  Link[]  = L"\\DosDevices\\K";
    UNICODE_STRING         uniLink;


    DbgPrint(("K: Treibereintrittspunkt\n"));

    //
    // Stop - definiertes Anhalten im Debugger - durch bedingte
    // Compilierung nur in der Debugversion.
    //
    
    DbgBreak();

    //
    // Device Object erzeugen
    //
	    
    RtlInitUnicodeString(&uniName,Name);
    mStatus=IoCreateDevice(DriverObject,0,&uniName,FILE_DEVICE_UNKNOWN,
                           0,TRUE,&DeviceObject );

    if (NT_SUCCESS(mStatus))
        {
	 //
	 // Erzeugen der symbolischen Verknpfung
         //

	 RtlInitUnicodeString(&uniLink,Link);
	 mStatus=IoCreateSymbolicLink(&uniLink,&uniName);
	 if (!NT_SUCCESS(mStatus))
	      DbgPrint(("K: IoCreateSymbolicLink - Fehler\n")); else
	      DbgPrint(("K: IoCreateSymbolicLink - Erfolgreich \n"));

	 //
	 // Einsprungpunkte fuer alle IRPs die der Tastaturtreiber
         // abarbeitet
         //

 	 DriverObject->MajorFunction[IRP_MJ_READ]	    = K_DispatchRead;
 	 
	 DriverObject->MajorFunction[IRP_MJ_CREATE]         =
	 DriverObject->MajorFunction[IRP_MJ_CLOSE]          =
	 DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS]  =
	 DriverObject->MajorFunction[IRP_MJ_CLEANUP]        =
	 DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = K_DispatchGeneral;
        }

    if (!NT_SUCCESS(mStatus))
        {
	 //
	 // Ein Fehler ist aufgetreten - Aufraeumarbeiten (Freigabe der belegten
	 // Ressourcen usw.)
	 //
	 
         DbgPrint(("K: ein Fehler ist aufgetreten - raeume auf ...\n"));
         if (DeviceObject) IoDeleteDevice(DeviceObject);
         return mStatus;
        }
    //
    // K_init aufrufen und den Rckgabewert an den I/O Manager zurck geben
    //    
    return K_Init(DriverObject);
}


//----------------------------------------------------------------------
//
// K_Init
//
// stellt die Verbindung zum Tastaturtreiber mit Hilfe der
// Funktion IoAttachDevice her
// 
//----------------------------------------------------------------------
NTSTATUS K_Init(IN PDRIVER_OBJECT DriverObject)
{
    CCHAR		 LowerDriverNameBuffer[64];
    STRING		 LowerDriverNameString;
    UNICODE_STRING       LowerDriverUnicodeString;
    NTSTATUS             mStatus;
    UCHAR                EventMsg[10];
    
    //
    // Umwandlung des Ansi Namens in Unicode
    //
    
    sprintf(LowerDriverNameBuffer,"\\Device\\KeyboardClass0");
    RtlInitAnsiString(&LowerDriverNameString,LowerDriverNameBuffer );
    RtlAnsiStringToUnicodeString(&LowerDriverUnicodeString,
                                 &LowerDriverNameString, TRUE );

    //
    // Erzeugen eines eigenen Device Objects fuer K
    //

    mStatus=IoCreateDevice(DriverObject,0,NULL,FILE_DEVICE_KEYBOARD,0,
			   FALSE,&HookDeviceObject );

    if (!NT_SUCCESS(mStatus))
        {
	 DbgPrint(("K: Fehler beim Erzeugen des Device Objects\n"));
	 RtlFreeUnicodeString( &LowerDriverUnicodeString );
	 return STATUS_SUCCESS;
        } else DbgPrint(("K: Device Object wurde erzeugt \n"));
   
    HookDeviceObject->Flags |= DO_BUFFERED_IO;
    
    //
    // Verbinden des eigenen Device Objects mit dem tieferliegenden
    // Treiber, dessen Name in LowerDriverUnicodeString steht.
    //
    // kbdDevice zeigt bei Erfolg auf das Device Object des
    // tieferliegenden Treibers.
    //
    
    mStatus=IoAttachDevice(HookDeviceObject,&LowerDriverUnicodeString,
                           &kbdDevice);

    if( !NT_SUCCESS(mStatus) )
        {
	 DbgPrint(("K: Verbindung zur Tastatur fehlgeschlagen !\n"));
	 IoDeleteDevice(HookDeviceObject);
	 RtlFreeUnicodeString(&LowerDriverUnicodeString);
	 return STATUS_UNSUCCESSFUL;
	} else DbgPrint(("K: Verbindung mit Tastatur hergestellt !\n"));

    RtlFreeUnicodeString(&LowerDriverUnicodeString);
    DbgPrint(("K: Erfolgreich initialisiert !\n"));

    K_ReportEvent(K_MSG_DRIVER_STARTING,1,(PVOID)DriverObject,NULL,NULL,0);
    sprintf(EventMsg,"%x",*kbdDevice);
    K_ReportEvent(K_MSG_DEVOBJ_ADR,2,(PVOID)DriverObject,NULL,EventMsg,10*sizeof(UCHAR));

    return STATUS_SUCCESS;
}

//----------------------------------------------------------------------
// 
// K_ReadComplete
//
// wird nach einer abgeschlossenen Leseoperation des tieferliegenden
// Treibers aufgerufen
//
//----------------------------------------------------------------------
NTSTATUS K_ReadComplete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp,
			IN PVOID Context)
{
    PIO_STACK_LOCATION        IrpSp;
    PKEYBOARD_INPUT_DATA      KeyData;
    int                       numKeys,i;

    IrpSp=IoGetCurrentIrpStackLocation(Irp);
    if (NT_SUCCESS(Irp->IoStatus.Status))
       {
        //
        // KeyData[].MakeCode = ScanCode
        // KeyData[].Flags = 0 - Taste gedrckt
        // KeyData[].Flags = 1 - Taste losgelassen
        //

  	KeyData = Irp->AssociatedIrp.SystemBuffer;
        numKeys = Irp->IoStatus.Information / sizeof(KEYBOARD_INPUT_DATA);

        //
        // Ausgabe aller Tastencodes auf dem Debugbildschirm
        //
        
        for (i=0;i<numKeys;i++)
           {
            DbgPrint(("K: ScanCode: %x ", KeyData[i].MakeCode ));
            DbgPrint((" %s\n", KeyData[i].Flags ? "Up" : "Down" ));
           }
            
	if ((KeyData[0].MakeCode==F12)&&(KeyData[0].Flags)&&(numKeys==1))
	   {
            DbgPrint(("K: simuliere Ctrl+Alt+Entf \n"));
            K_ReportEvent(K_MSG_SIM_CTRL_ALT_ENTF,3,(PVOID)DeviceObject,Irp,NULL,0);
                
	    KeyData[0].MakeCode = CTRL;
	    KeyData[0].Flags    = KEY_DOWN;
	    KeyData[1].MakeCode = ALT;
	    KeyData[1].Flags    = KEY_DOWN;
	    KeyData[2].MakeCode = ENTF;
	    KeyData[2].Flags    = KEY_DOWN;

	    KeyData[3].MakeCode = CTRL;
	    KeyData[3].Flags    = KEY_UP;
	    KeyData[4].MakeCode = ALT;
	    KeyData[4].Flags    = KEY_UP;
	    KeyData[5].MakeCode = ENTF;
	    KeyData[5].Flags    = KEY_UP;
	       
	    Irp->IoStatus.Information=6*sizeof(KEYBOARD_INPUT_DATA);
           }
       }

    if (Irp->PendingReturned) IoMarkIrpPending(Irp);
    return Irp->IoStatus.Status;
}


//----------------------------------------------------------------------
//
// K_DispatchRead
//
// setzt Rueckruffunktion und ruft den unteren Tastaturtreiber
//
//----------------------------------------------------------------------
NTSTATUS K_DispatchRead( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
    PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    PIO_STACK_LOCATION nextIrpStack    = IoGetNextIrpStackLocation(Irp);

    *nextIrpStack = *currentIrpStack;
    IoSetCompletionRoutine(Irp,K_ReadComplete,DeviceObject,TRUE,TRUE,TRUE );

    //
    // Rueckgabewert = Rueckgabe des unteren Tastaturtreibers
    //
	    
    return IoCallDriver(kbdDevice,Irp);
}


//----------------------------------------------------------------------
//
// K_DispatchGeneral
//
// reicht alle Anfrage an den unteren Tastaturtreiber an diesen weiter,
// sonst Rueckgabe von STATUS_SUCCESS
//
//----------------------------------------------------------------------
NTSTATUS K_DispatchGeneral(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp )
{
    PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    PIO_STACK_LOCATION nextIrpStack    = IoGetNextIrpStackLocation(Irp);

    Irp->IoStatus.Status      = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

   if( DeviceObject == HookDeviceObject )
      {
       *nextIrpStack = *currentIrpStack;
       return IoCallDriver( kbdDevice, Irp );
      } else return STATUS_SUCCESS;
}


//----------------------------------------------------------------------
//
// K_ReportEvent
//
// schreibt einen Eintrag in das Eventlog
//
//----------------------------------------------------------------------

BOOLEAN K_ReportEvent(IN NTSTATUS ErrorCode,IN ULONG UniqueErrorValue,
                      IN PVOID IoObject,IN PIRP Irp,PUCHAR InsertString,
                      IN ULONG StringSize)
{
	PIO_ERROR_LOG_PACKET Packet;
	PIO_STACK_LOCATION IrpStack;
	PWCHAR pInsertionString;
	STRING AnsiInsertString;
        UNICODE_STRING UniInsertString;

	UCHAR PacketSize;


	PacketSize = sizeof(IO_ERROR_LOG_PACKET) + (StringSize+1)*sizeof(WCHAR);

	Packet = IoAllocateErrorLogEntry(IoObject,PacketSize);
	if (Packet == NULL) return FALSE;

	Packet->ErrorCode         = ErrorCode;
	Packet->UniqueErrorValue  = UniqueErrorValue;
	Packet->RetryCount        = 0;
	Packet->SequenceNumber    = 0;
	Packet->IoControlCode     = 0;

//	
// K benutzt das DumpData Feld nicht, daher DumpDataSize=0
//

	Packet->DumpDataSize      = 0;
	
	if (Irp!=NULL)
	  {
	   IrpStack=IoGetCurrentIrpStackLocation(Irp);
	   Packet->MajorFunctionCode = IrpStack->MajorFunction;
	   Packet->FinalStatus = Irp->IoStatus.Status;
	  } else
	  {
           Packet->MajorFunctionCode = 0;
           Packet->FinalStatus       = 0;
	  }
//
// Konvertierung Ansi - Unicode nur im PASSIVE_LEVEL moeglich
//
        if ((StringSize>0)&&(KeGetCurrentIrql()==PASSIVE_LEVEL))
	  {
	   Packet->NumberOfStrings=1;
	   Packet->StringOffset=sizeof(IO_ERROR_LOG_PACKET);
	   RtlInitAnsiString(&AnsiInsertString,InsertString);
	   RtlAnsiStringToUnicodeString(&UniInsertString,&AnsiInsertString,TRUE);
	   pInsertionString = (PUCHAR)Packet + Packet->StringOffset;
           RtlCopyBytes(pInsertionString,UniInsertString.Buffer,
                        UniInsertString.Length+sizeof(WCHAR));
           RtlFreeUnicodeString(&UniInsertString);
          } else Packet->NumberOfStrings=0;

	IoWriteErrorLogEntry(Packet);
	return TRUE;
}
